package dev.pfaff.altargui.render;

import dev.pfaff.glass.Interop;
import dev.pfaff.glass.opengl.ConstPointer;
import dev.pfaff.glass.opengl.enums.GLDebugSeverity;
import dev.pfaff.glass.opengl.enums.GLDebugSource;
import dev.pfaff.glass.opengl.enums.GLDebugType;
import dev.pfaff.log4truth.Logger;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.Arena;
import java.lang.foreign.FunctionDescriptor;
import java.lang.foreign.Linker;
import java.lang.foreign.MemorySegment;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.util.List;

import static dev.pfaff.altargui.internal.InternalGlobals.GL_FEATURES;
import static java.lang.invoke.MethodHandles.filterArguments;
import static java.lang.invoke.MethodType.methodType;
import static java.util.FormatProcessor.FMT;

public final class GLDebugCallback {
	static {
		var khrDebug = GL_FEATURES.getKHR_Debug();
		if (khrDebug != null) {
			MethodHandle mh;
			try {
				mh = MethodHandles.lookup().findStatic(GLDebugCallback.class, "handleMessage",
													   methodType(void.class,
																  int.class,
																  int.class,
																  int.class,
																  int.class,
																  long.class,
																  MemorySegment.class,
																  MemorySegment.class));
			} catch (NoSuchMethodException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}
			mh = filterArguments(mh, 4, Interop.MH_cast_usize_to_long);
			var fd = FunctionDescriptor.ofVoid();
			var stub = Linker.nativeLinker().upcallStub(mh, fd, Arena.global());
			khrDebug.glDebugMessageCallback(new ConstPointer<>(stub), null);
		}
	}

	public static void register() {}

	private static final List<List<String>> TAGS_MATRIX;

	static {
		assert GLDebugSource.DONT_CARE.ordinal() == 0;
		assert GLDebugType.DONT_CARE.ordinal() == 0;
		assert GLDebugSeverity.DONT_CARE.ordinal() == 0;
		TAGS_MATRIX = GLDebugSource.VALUES.stream()
			.filter(source -> source != GLDebugSource.DONT_CARE)
			.map(source -> "GL" + source)
			.flatMap(source -> GLDebugType.VALUES.stream()
				.filter(type -> type != GLDebugType.DONT_CARE)
				.map(type -> "GL" + type)
				.flatMap(type -> GLDebugSeverity.VALUES.stream()
					.filter(severity -> severity != GLDebugSeverity.DONT_CARE)
					.map(severity -> List.of(source, type, "GL" + severity))))
			.toList();
	}

	private static void handleMessage(GLDebugSource source,
							  GLDebugType type,
							  int id,
							  GLDebugSeverity severity,
							  long length,
							  ConstPointer<Byte> message,
							  @Nullable ConstPointer<Void> userParam) {
		assert userParam == null;
		var tagsMatrixIndex = (source.ordinal() - 1) * (GLDebugType.VALUES.size() - 1) * (GLDebugSeverity.VALUES.size() - 1)
			+ (type.ordinal() - 1) * (GLDebugSeverity.VALUES.size() - 1)
			+ (severity.ordinal() - 1);
		Logger.log(() -> {
			return FMT."0x%08x\{id}: \{message.ms().reinterpret(length + 1).getString(0)}";
		}, TAGS_MATRIX.get(tagsMatrixIndex));
	}

	private static void handleMessage(int source,
							  int type,
							  int id,
							  int severity,
							  long length,
							  MemorySegment message,
							  MemorySegment userParam) {
		handleMessage(GLDebugSource.fromGLEnum(source),
					  GLDebugType.fromGLEnum(type),
					  id,
					  GLDebugSeverity.fromGLEnum(severity),
					  length,
					  new ConstPointer<>(message),
					  ConstPointer.newNullable(userParam));
	}
}
